from typing import List, Tuple, Dict
import numpy as np
from numpy.lib import recfunctions as rfn
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda


NAME = "ROE"
DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "ceq",
            "ib",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    )
]
VARIABLE_LABELS = {
    NAME: "Income Before Extraordinary Items scaled by Common Equity (Total)"
}


class ROE(Measure):
    """ROE:
        Income Before Extraordinary Items scaled by Assets (Total)
     =  ----------------------------------------------------------
        Common Equity
    """

    def __init__(self):
        super().__init__("ROE", DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):
        nparray = filter_funda(nparrays[0])
        roa = np.true_divide(nparray.ib, nparray.ceq, where=(nparray.ceq != 0))
        roa[np.isnan(nparray.ceq)] = np.nan
        nparray = rfn.rec_append_fields(nparray, NAME, roa)
        # keep only useful columns
        cols = set(rfn.get_names_flat(nparray.dtype))
        nparray.sort(order=(keys := ["gvkey", "datadate"]))
        exclude_cols = cols - set([*keys, "ib", "ceq", NAME])
        return (
            pd.DataFrame.from_records(nparray, exclude=exclude_cols),
            VARIABLE_LABELS,
        )



from typing import List, Tuple, Dict
import numpy as np
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda

NAME = "ExecutiveOwnership"
DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "fyear",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
            "csho",
        ],
        date_vars=["datadate"],
    ),
    Dataset(
        source="wrds",
        library="execcomp",
        table="anncomp",
        vars=[
            "gvkey",
            "year",
            "execid",
            "shrown_tot",
            "shrown_excl_opts",
            "opt_exer_num",
            "opt_exer_val",
            "shrs_vest_num",
            "shrs_vest_val",
            "tdc1",
        ],
        date_vars=[],
    ),
]

VARIABLE_LABELS: Dict[str, str] = {
    "execid": "Executive ID from Execucomp",
    "ExecSharePct": "Executive share ownership (%)",
    "ExecSharePctExclOpt": "Executive share ownership (%) excluding options",
    "ExecOptPct": "Executive share ownership (%) based on shares acquired on option exercise",
    "ExecShareVestPct": "Executive share ownership (%) based on shared acquired on vesting",
    "ExecIncentivePct": "Value realized on option exercise and vesting scaled by total compensation",
}


class ExecutiveOwnership(Measure):
    """Compute various executive ownership measures"""

    def __init__(self):
        super().__init__("Executive Ownership", DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):

        funda = pd.DataFrame.from_records(filter_funda(nparrays[0]))
        anncomp = pd.DataFrame.from_records(nparrays[1])

        work = anncomp.merge(
            funda, left_on=["gvkey", "year"], right_on=["gvkey", "fyear"]
        )

        # CSHO is in millions and SHROWN_TOT is in thousands
        work["ExecSharePct"] = work.shrown_tot / work.csho / 10
        work["ExecSharePctExclOpt"] = work.shrown_excl_opts / work.csho / 10
        work["ExecOptPct"] = work.opt_exer_num / work.csho / 10
        work["ExecShareVestPct"] = work.shrs_vest_num / work.csho / 10
        work["ExecIncentivePct"] = (
            (work.opt_exer_val + work.shrs_vest_num) / work.tdc1 / 10
        )
        # Replace infinity with nan
        work["ExecSharePctExclOpt"].replace(np.inf, np.nan, inplace=True)
        work["ExecOptPct"].replace(np.inf, np.nan, inplace=True)
        work["ExecShareVestPct"].replace(np.inf, np.nan, inplace=True)
        work["ExecIncentivePct"].replace(np.inf, np.nan, inplace=True)
        keys = ["gvkey", "datadate"]
        cols = [*keys, *(VARIABLE_LABELS.keys())]
        result = work[cols].drop_duplicates().sort_values(by=keys)
        return result, VARIABLE_LABELS



"""DataManager manages the shared memory"""

import os
from typing import List, Dict
from multiprocessing.shared_memory import SharedMemory
from multiprocessing.managers import SharedMemoryManager
from importlib import import_module
import numpy as np
import pandas as pd
from frds.typing import TableID, SharedMemoryInfo
from .dataset import Dataset


class Singleton(object):
    """Singleton implementation by Guido van Rossum"""

    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        pass


class DataManager(Singleton):
    """DataManager loads data from sources and manages the shared memory"""

    def __init__(self, obs=-1, config=None):
        self._datasets = dict()
        self.smm = SharedMemoryManager()
        self.smm.start()
        self.conns = dict()
        self.obs = obs
        self.config = config

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    def shutdown(self):
        """Shutdown the shared memory manager and all data connections"""
        self.smm.shutdown()
        for conn in self.conns.values():
            conn.close()

    def _add_to_shared_memory(self, nparray: np.recarray) -> SharedMemory:
        """Internal function to copy an array into shared memory.

        Parameters
        ----------
        nparray : np.recarray
            The array to be copied into shared memory.

        Returns
        -------
        SharedMemory
            The shared memory object.
        """
        shm = self.smm.SharedMemory(nparray.nbytes)
        array = np.recarray(nparray.shape, dtype=nparray.dtype, buf=shm.buf)
        np.copyto(array, nparray)
        return shm

    def _download_dataset(self, dataset: Dataset) -> pd.DataFrame:
        """Internal function to download the dataset.

        Parameters
        ----------
        dataset : Dataset
            Dataset information including source, library, table, vars, etc.

        Returns
        -------
        pd.DataFrame
            DataFrame of the downloaded dataset.
        """
        # TODO: generic login data for different data sources
        if dataset.source == "wrds":
            usr = self.config.get("wrds_username")
            pwd = self.config.get("wrds_password")
        # If there exists a connection for the data source, use it!
        if (conn := self.conns.get(dataset.source, None)) is None:
            module = import_module(f"frds.data.{dataset.source}")
            conn = module.Connection(usr=usr, pwd=pwd)
            self.conns.update({dataset.source: conn})
        return conn.get_table(
            library=dataset.library,
            table=dataset.table,
            columns=dataset.vars,
            date_cols=dataset.date_vars,
            obs=self.obs,
        )

    def _get_dataset(self, dataset: Dataset) -> SharedMemoryInfo:
        """Get the shared memory object, shape and dtype of the given dataset.

        Parameters
        ----------
        dataset : Dataset
            Dataset information including source, library, table, vars, etc.

        Returns
        -------
        SharedMemoryInfo
            (SharedMemory, tuple, numpy.dtype), where the tuple is the shape.
        """
        if dataset in self._datasets:
            return self._datasets.get(dataset)
        dataset_name = f"{dataset.source}_{dataset.library}_{dataset.table}"
        path = os.path.join(self.config.get("data_dir"), f"{dataset_name}.dta")
        # If local data directory has the table
        if os.path.exists(path):
            df = next(pd.read_stata(path, chunksize=1))
            # If required variables are all in the table
            if set(dataset.vars).issubset(set(df.columns)):
                df = pd.read_stata(path, columns=dataset.vars)
            # Some variables are missing and need to be downloaded
            else:
                # Should find out the missing variables and only download them
                # But since we don't know the identifiers, might as well just
                # download everything and merge with existing one.
                df = self._download_dataset(dataset)
                df_in_datadir = pd.read_stata(path)
                common_cols = [
                    col for col in df.columns if col in df_in_datadir.columns
                ]
                df_to_save = df.merge(df_in_datadir, on=common_cols)
                df_to_save.to_stata(
                    path,
                    write_index=False,
                    convert_dates={v: "td" for v in dataset.date_vars},
                )
                del df_in_datadir
                del df_to_save
        # No such table in the local data directory
        else:
            # Download the entire table from source
            df = self._download_dataset(dataset)
            # Store the table in local data directory
            df.to_stata(
                path,
                write_index=False,
                convert_dates={v: "td" for v in dataset.date_vars},
            )
        # VERY IMPORTANT: must find out the string-typed columns
        dtypes = df.convert_dtypes().dtypes
        str_cols = dtypes.where(dtypes == "string").dropna().index.to_list()
        # Stata variable name's length is maxed at 32. Here I use 32 characters
        # unicode dtype because "S32" maps to zero-terminated bytes `np.bytes_`
        # for backward compatibility in Python 2. In Python 3 we use unicode.
        # This will slightly increase the array size, but should be no harm.
        nparray = df.to_records(
            index=False, column_dtypes={col: "U32" for col in str_cols}
        )
        shm = self._add_to_shared_memory(nparray)
        self._datasets.update({dataset: (shm, nparray.shape, nparray.dtype)})
        return shm, nparray.shape, nparray.dtype

    def get_datasets(
        self, datasets: List[Dataset]
    ) -> Dict[TableID, SharedMemoryInfo]:
        """Return a dictionary of {TableID: SharedMemoryInfo}.

        Parameters
        ----------
        datasets : List[Dataset]
            List of requried datasets.

        Returns
        -------
        Dict[TableID, SharedMemoryInfo]
            Dictionary of TableID to SharedMemoryInfo.
        """
        return {dta.table_id: self._get_dataset(dta) for dta in datasets}




from typing import List, Tuple, Dict
import numpy as np
from numpy.lib import recfunctions as rfn
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda


NAME = "AssetTangibility"
DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "at",
            "ppent",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    )
]
VARIABLE_LABELS = {
    NAME: "Property, Plant and Equipment (Net) scaled by Assets (Total)"
}


class AssetTangibility(Measure):
    """Asset tangibility:
        Plant, Property and Equipment (Net)
     =  -----------------------------------
        Assets (Total)
    """

    def __init__(self):
        super().__init__("Asset Tangibility", DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):
        nparray = filter_funda(nparrays[0])
        tangibility = np.true_divide(
            nparray.ppent, nparray.at, where=(nparray.at != 0)
        )
        tangibility[np.isnan(nparray.at)] = np.nan
        nparray = rfn.rec_append_fields(nparray, NAME, tangibility)
        # keep only useful columns
        cols = set(rfn.get_names_flat(nparray.dtype))
        nparray.sort(order=(keys := ["gvkey", "datadate"]))
        exclude_cols = cols - set([*keys, "ppent", "at", NAME])
        return (
            pd.DataFrame.from_records(nparray, exclude=exclude_cols),
            VARIABLE_LABELS,
        )



"""Defines the base class for all measures"""
import abc
from typing import List, Tuple
import numpy as np
import pandas as pd
from frds.data import Dataset
from frds.typing import TableID


class Measure(abc.ABC):
    """Base class for all measures"""

    def __init__(self, name: str, datasets_required: List[Dataset]):
        self._name = name
        self._datasets_required = datasets_required

    @property
    def name(self) -> str:
        """Return the measure's name

        Returns
        -------
        str
            Name of the measure
        """
        return self._name

    def __str__(self) -> str:
        """Return the measure's name by default. However, actual measures can
        choose to overwrite this if they have custom parameters necessary to
        identify the measure

        Returns
        -------
        str
            Name of the measure
        """
        return self._name

    @property
    def datasets_required(self) -> List[Dataset]:
        """Return the measure's required datasets

        Returns
        -------
        List[Dataset]
            List of required `frds.data.Dataset`s
        """
        return self._datasets_required

    @property
    def data_sources(self) -> List[str]:
        """Return the list of unique data sources required

        Returns
        -------
        List[str]
            List of required data sources
        """
        return list(set(dta.source for dta in self._datasets_required))

    @property
    def data_tables(self) -> List[TableID]:
        """Return the list of unique tables required

        Returns
        -------
        list[TableID]
            List of (dataset.source, dataset.library, dataset.table)
        """
        return list(set(dta.table_id for dta in self._datasets_required))

    @abc.abstractmethod
    def estimate(
        self, nparrays: List[np.recarray]
    ) -> Tuple[pd.DataFrame, dict]:
        """Estimatation function for the measure to be implemented

        Parameters
        ----------
        nparrays : List[np.recarray]
            List of `numpy.recarray`s corresponding to the required datasets

        Returns
        -------
        Tuple[pd.DataFrame, dict]
            Tuple of result `pandas.DataFrame` and dictionay of variable labels
        """
        raise NotImplementedError



from .base import Measure
from .accounting_restatement import AccountingRestatement
from .book_leverage import BookLeverage
from .capital_expenditure import CapitalExpenditure
from .executive_ownership import ExecutiveOwnership
from .market_to_book import MarketToBookRatio
from .roa import ROA
from .roe import ROE
from .firm_size import FirmSize
from .asset_tangibility import AssetTangibility
from .board_independence import BoardIndependence



"""Define the research assistant process"""

import os
from multiprocessing import Process
from multiprocessing.managers import Namespace
from typing import Dict
from numpy import recarray
from pandas import DataFrame
from frds.measures import Measure
from frds.typing import TableID, SharedMemoryInfo


class ResearchAssistant(Process):
    """Research Assistant is the worker (process) who manages and performs the
    estimation, as well as saving results.

    Parameters
    ----------
    measure : Measure
        The measure to estimate.
    
    data : Dict[TableID, SharedMemoryInfo]
        Used to know where to locate the required datasets in memory
    """

    def __init__(self, measure: Measure, office: Namespace):
        super(ResearchAssistant, self).__init__()
        self.measure = measure
        self.data: Dict[TableID, SharedMemoryInfo] = office.data
        self.config = office.config

    def run(self):
        # Load data from shared memory
        datasets = []
        for dta in self.measure.datasets_required:
            shm, shape, dtype = self.data.get(dta.table_id)
            datasets.append(recarray(shape, dtype=dtype, buf=shm.buf))
        # Invoke the estimation function of the measure
        result, variable_labels = self.measure.estimate(datasets)
        self._save_results(result, variable_labels)

    def _save_results(self, result: DataFrame, variable_labels: Dict[str, str]):
        """Save results to local result directory

        RAs can save results themselves. No need to return to main process

        Parameters
        ----------
        result : DataFrame
            Result dataset
        variable_labels : Dict[str, str]
            Dictionary of variable name to label
        """
        assert isinstance(result, DataFrame)
        # Find out the date variables
        date_vars = set()
        for dta in self.measure.datasets_required:
            date_vars.update(dta.date_vars)
        # Save to the result directory
        result_dir = self.config.get("result_dir")
        stata_file = f"{self.measure}.dta"
        stata_path = os.path.join(result_dir, stata_file)
        result.to_stata(
            path=stata_path,
            write_index=False,
            convert_dates={v: "td" for v in date_vars if v in result.columns},
            variable_labels=variable_labels,
        )




from typing import List, Tuple, Dict
import numpy as np
from numpy.lib import recfunctions as rfn
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda

NAME = "MarketToBookRatio"
DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "csho",
            "prcc_f",
            "ceq",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    )
]
VARIABLE_LABELS = {NAME: "(PRCC_F*CSHO)/CEQ"}


class MarketToBookRatio(Measure):
    """Market to book ratio
        common shares outstanding * share price at fiscal year end
     =  ----------------------------------------------------------
        book value common equity
    """

    def __init__(self):
        super().__init__("Market to Book Ratio", DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):
        nparray = filter_funda(nparrays[0])
        # market value at fiscal year
        mv = nparray.prcc_f * nparray.csho
        # market-to-book = market value of equity / common equity
        mtb = np.true_divide(mv, nparray.ceq, where=(nparray.ceq != 0))
        # set mtb to missing if common equity is somehow missing
        mtb[np.isnan(nparray.ceq)] = np.nan
        # add book leverage to the result
        nparray = rfn.rec_append_fields(nparray, NAME, mtb)
        # keep only useful columns
        cols = set(rfn.get_names_flat(nparray.dtype))
        nparray.sort(order=(keys := ["gvkey", "datadate"]))
        exclude_cols = cols - set([*keys, "prcc_f", "csho", "ceq", NAME])
        return (
            pd.DataFrame.from_records(nparray, exclude=exclude_cols),
            VARIABLE_LABELS,
        )



from typing import List, Tuple, Dict
import numpy as np
from numpy.lib import recfunctions as rfn
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda


NAME = "FirmSize"
DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "at",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    )
]
VARIABLE_LABELS = {NAME: "Natural logarithm of total assets"}


class FirmSize(Measure):
    """Firm size: the natural logarithm of total assets"""

    def __init__(self):
        super().__init__("Firm Size", DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):
        nparray = filter_funda(nparrays[0])
        size = np.log(nparray.at, where=(nparray.at > 0))
        size[np.isnan(nparray.at)] = np.nan
        nparray = rfn.rec_append_fields(nparray, NAME, size)
        # keep only useful columns
        cols = set(rfn.get_names_flat(nparray.dtype))
        nparray.sort(order=(keys := ["gvkey", "datadate"]))
        exclude_cols = cols - set([*keys, "at", NAME])
        return (
            pd.DataFrame.from_records(nparray, exclude=exclude_cols),
            VARIABLE_LABELS,
        )



import os
import sys
import pathlib
from configparser import ConfigParser, ExtendedInterpolation
from .professor import Professor

if sys.version_info.major < 3 and sys.version_info.minor < 8:
    print("Python3.8 and higher is required.")
    sys.exit(1)

config = ConfigParser(interpolation=ExtendedInterpolation())

# custom config.ini in user's directory
custom_config = os.path.join(pathlib.Path("~/frds").expanduser(), "config.ini")
config.read(custom_config)

data_dir = pathlib.Path(config["Paths"]["data_dir"]).expanduser()
result_dir = pathlib.Path(config["Paths"]["result_dir"]).expanduser()

os.makedirs(data_dir, exist_ok=True)
os.makedirs(result_dir, exist_ok=True)

conf = config["Login"]
wrds_username = conf.get("wrds_username", fallback=None)
wrds_password = conf.get("wrds_password", fallback=None)

if not all([wrds_username, wrds_password]):
    print(f"Please enter your WRDS username and password in {custom_config}.")



from typing import List, Tuple, Dict
import numpy as np
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda

DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="boardex",
        table="na_wrds_company_profile",
        vars=["cikcode", "boardid"],
        date_vars=[],
    ),
    Dataset(
        source="wrds",
        library="boardex",
        table="na_wrds_org_composition",
        vars=[
            "companyid",
            "datestartrole",
            "dateendrole",
            "rolename",
            "directorid",
            "seniority",
        ],
        date_vars=["datestartrole", "dateendrole"],
    ),
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "fyear",
            "cik",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    ),
]


VARIABLE_LABELS: Dict[str, str] = {
    "BoardSize": "Number of directors",
    "IndependentMembers": "Number of independent members on board",
    "BoardIndependence": "Ratio of independent board members to board size (%)",
}


class BoardIndependence(Measure):
    def __init__(self, missing_independent_board_members_as_zero=True):
        super().__init__("Board Size and Independence", DATASETS_REQUIRED)
        self._missing_as_zero = missing_independent_board_members_as_zero

    def estimate(
        self, nparrays: List[np.recarray]
    ) -> Tuple[pd.DataFrame, Dict[str, str]]:
        company = pd.DataFrame.from_records(nparrays[0])
        compositon = pd.DataFrame.from_records(nparrays[1])
        funda = pd.DataFrame.from_records(
            filter_funda(nparrays[2]),
            columns=["gvkey", "datadate", "cik", "fyear"],
        )
        tmp = (
            company.merge(
                compositon, left_on=["boardid"], right_on=["companyid"]
            )
            .merge(funda, left_on=["cikcode"], right_on=["cik"])
            .drop_duplicates()
        )
        cond = (
            (
                (tmp.datestartrole <= tmp.datadate)
                & (tmp.datadate <= tmp.dateendrole)
            )
            | (pd.isnull(tmp.dateendrole) & (tmp.datestartrole <= tmp.datadate))
            | (pd.isnull(tmp.datestartrole) & (tmp.datadate <= tmp.dateendrole))
        ) & np.in1d(
            tmp.seniority, ("Executive Director", "Supervisory Director")
        )
        tmp = tmp.where(cond)
        tmp.fillna({"rolename": ""}, inplace=True)
        keys = ["gvkey", "datadate"]
        board_size = (
            tmp.groupby(keys, as_index=False)["directorid"]
            .count()
            .rename(columns={"directorid": "BoardSize"})
        )
        independent_dirs = (
            tmp[tmp.rolename.str.lower().str.contains("independent")]
            .groupby(keys, as_index=False)["directorid"]
            .count()
            .rename(columns={"directorid": "IndependentMembers"})
        )

        result = board_size.merge(independent_dirs, on=keys, how="left")

        if self._missing_as_zero:
            result.fillna({"IndependentMembers": 0}, inplace=True)

        result["BoardIndependence"] = (
            result.IndependentMembers / result.BoardSize * 100
        )

        cols = [*keys, *(VARIABLE_LABELS.keys())]
        result = result[cols].drop_duplicates()

        return result, VARIABLE_LABELS



"""
This measure counts the numer of restatements during the past fiscal year(s).

Data Souces
-----------
WRDS Compustat Fundamentals Annual, AuditAnalytics Non-Reliance Restatement

Author
------
Mingze Gao, 16 June 2020
"""
from typing import List, Tuple, Dict
import numpy as np
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda

DATASETS_REQUIRED: List[Dataset] = [
    Dataset(
        source="wrds",
        library="audit",
        table="auditnonreli",
        vars=[
            "company_fkey",  # EDGAR CIK
            "file_date",  # Filing date
            "res_notif_key",  # Restatement notification key
            "res_accounting",  # Restatement accounting
            "res_adverse",  # Restatement adverse
            "res_fraud",  # Restatement fraud
            "res_cler_err",  # Restatement clerical errors
            "res_sec_invest",  # Restatement SEC investigation
        ],
        date_vars=["file_date"],
    ),
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "gvkey",
            "datadate",
            "cik",
            "datafmt",
            "indfmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    ),
]
"""List[Dataset]: list of `frds.data.Dataset` required for this measure"""

VARIABLE_LABELS: Dict[str, str] = {
    "NumResAcct": "Numer of restatements (accounting) in the fiscal year",
    "NumResFraud": "Numer of restatements (fraud) in the fiscal year",
    "NumResAdver": "Numer of restatements (adverse) in the fiscal year",
    "NumResClerErr": "Numer of restatements (clerical errors) in the fiscal year",
    "NumResSECInvest": "Numer of restatements (SEC investigation) in the fiscal year",
}
"""Dict[str, str]: dict of the variable labels for the output Stata file"""


class AccountingRestatement(Measure):
    """Counts the numer of restatements during the past fiscal year"""

    def __init__(self, years=1):
        """Set numer of fiscal years to consider

        Parameters
        ----------
        years : int, optional
            Number of past fiscal years to consider, by default 1
        """
        super().__init__(
            name="Accounting Restatement", datasets_required=DATASETS_REQUIRED
        )
        self._years = int(years)

    def __str__(self):
        """Return the measure's full name, including the parameters used

        Returns
        -------
        str
            Measure's full name, including the parameters (years used)
        """
        return f"{self.name} years={self._years}"

    def estimate(
        self, nparrays: List[np.recarray]
    ) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """Counts the numer of restatements during the past fiscal year(s)

        Parameters
        ----------
        nparrays : List[np.recarray]
            Input datasets, same order as specified in `DATASETS_REQUIRED`

        Returns
        -------
        Tuple[pd.DataFrame, Dict[str, str]]
            Output dataset and the `VARIABLE_LABELS`
        """
        # Note that the order of dataset is preseved
        rests = pd.DataFrame.from_records(nparrays[0])
        funda = pd.DataFrame.from_records(filter_funda(nparrays[1]))
        # Inner join funda and restatements
        work = funda.merge(rests, left_on="cik", right_on="company_fkey")
        # Filing date must be in the past year(s) relative the datadate
        time_delta = work.datadate - work.file_date
        years = np.timedelta64(self._years, "Y")
        cond = (time_delta < years) & (work.datadate >= work.file_date)
        # Forget about the non-matched
        cond = cond & (work.cik != "")
        # Apply the filtering condition
        work.where(cond, inplace=True)
        # Count by gvkey and datadate/fyear
        keys = ["gvkey", "datadate"]
        work = work.groupby((keys), as_index=False).sum()
        # Rename columns to match output variable labels
        work.rename(
            columns={
                "res_accounting": "NumResAcct",
                "res_fraud": "NumResFraud",
                "res_adverse": "NumResAdver",
                "res_cler_err": "NumResClerErr",
                "res_sec_invest": "NumResSECInvest",
            },
            inplace=True,
        )
        # Left join with funda so to retain the missing values
        result = funda[keys].merge(work, how="left", on=keys, copy=False)
        # Keep only useful columns
        cols = [*keys, *(VARIABLE_LABELS.keys())]
        # Some cosmetic issues
        result = result[cols].drop_duplicates().sort_values(by=keys)

        return result, VARIABLE_LABELS



import unittest
import numpy as np
from frds.data import wrds, DataManager
from frds.measures import sample_measure


class Login(unittest.TestCase):

    def test_login_with_empty_username_and_password(self):
        with self.assertRaises(Exception):
            wrds.Connection(usr=None, pwd=None)


class GetData(unittest.TestCase):

    def test_get_data(self):
        with DataManager(obs=100) as dm:
            # Should be able to create np.recarry from shared memory
            shm, shape, dtype = dm.get_dataset(sample_measure.datasets[0])
            nparray = np.recarray(shape=shape, dtype=dtype, buf=shm.buf)
            self.assertIsInstance(nparray, np.recarray)
            self.assertGreater(np.nanmean(nparray.at), 0)
            # The same dataset should have the same shared memory name
            shm2, _, _ = dm.get_dataset(sample_measure.datasets[0])
            self.assertEqual(shm.name, shm2.name)
            self.assertEqual(len(dm._datasets), 1)



"""Define the base structure of a dataset"""
from dataclasses import dataclass
from typing import List
from frds.typing import TableID


@dataclass
class Dataset:
    """The base class of a dataset"""

    source: str
    library: str
    table: str
    vars: List[str]
    date_vars: List[str]

    def __hash__(self):
        return id(self)

    @property
    def table_id(self) -> TableID:
        """Return the tuple of (source, library, table)

        Returns
        -------
        TableID
            (source, library, table)
        """
        return TableID((self.source, self.library, self.table))



from typing import List, Tuple, Dict
import numpy as np
from numpy.lib import recfunctions as rfn
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda

NAME = "BookLeverage"
DATASETS_REQUIRED = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "dltt",
            "dlc",
            "ceq",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    )
]
VARIABLE_LABEL: Dict[str, str] = {NAME: "(DLTT+DLC)/(DLTT+DLC+CEQ)"}


class BookLeverage(Measure):
    """Book leverage:
        (long-term debt + debt in current liabilities) 
     =  --------------------------------------------------------------
        (long-term debt + debt in current liabilities + common equity)

    If CEQ is missing, the book leverage is treated as missing. 
    This avoids a book leverage of 100%."
    """

    def __init__(self):
        super().__init__("Book Leverage", datasets_required=DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):
        """Compute the book leverage

        Parameters
        ----------
        nparrays : List[np.recarray]
            Input datasets, same order as specified in `DATASETS_REQUIRED`

        Returns
        -------
        Tuple[pd.DataFrame, Dict[str, str]]
            Output dataset and the `VARIABLE_LABELS`
        """
        nparray = filter_funda(nparrays[0])
        # debts = long-term debt + debt in current liabilities
        debts = np.nansum([nparray.dltt, nparray.dlc], axis=0)
        # assets = debts + common equity
        assets = np.nansum([debts, nparray.ceq], axis=0)
        # book leverage = debts / assets
        bleverage = np.true_divide(debts, assets, where=(assets != 0))
        # set book leverage to missing if common equity is somehow missing
        bleverage[np.isnan(nparray.ceq)] = np.nan
        # add book leverage to the result
        nparray = rfn.rec_append_fields(nparray, NAME, bleverage)
        # keep only useful columns
        cols = set(rfn.get_names_flat(nparray.dtype))
        nparray.sort(order=(keys := ["gvkey", "datadate"]))
        exclude_cols = cols - set([*keys, "dltt", "dlc", "ceq", NAME])
        return (
            pd.DataFrame.from_records(nparray, exclude=exclude_cols),
            VARIABLE_LABEL,
        )



from typing import List, Tuple, Dict
import numpy as np
from numpy.lib import recfunctions as rfn
import pandas as pd
from frds.data import Dataset
from frds.measures import Measure
from frds.data.utils import filter_funda

NAME = "CapitalExpenditure"
DATASETS_REQUIRED = [
    Dataset(
        source="wrds",
        library="comp",
        table="funda",
        vars=[
            "datadate",
            "gvkey",
            "at",
            "capx",
            "indfmt",
            "datafmt",
            "popsrc",
            "consol",
        ],
        date_vars=["datadate"],
    )
]
VARIABLE_LABELS: Dict[str, str] = {
    NAME: "Capital Expenditures scaled by Assets (Total)"
}


class CapitalExpenditure(Measure):
    """Capial expenditure scaled by total assets

        capital expenditures
     =  --------------------
        total assets
    """

    def __init__(self):
        super().__init__("Capital Expendture", DATASETS_REQUIRED)

    def estimate(self, nparrays: List[np.recarray]):
        """Compute the capital expenditures scaled by total assets

        Parameters
        ----------
        nparrays : List[np.recarray]
            Input datasets, same order as specified in `DATASETS_REQUIRED`

        Returns
        -------
        Tuple[pd.DataFrame, Dict[str, str]]
            Output dataset and the `VARIABLE_LABELS`
        """
        nparray = filter_funda(nparrays[0])
        capx = np.true_divide(nparray.capx, nparray.at, where=(nparray.at != 0))
        capx[np.isnan(nparray.at)] = np.nan
        nparray = rfn.rec_append_fields(nparray, NAME, capx)
        # keep only useful columns
        cols = set(rfn.get_names_flat(nparray.dtype))
        nparray.sort(order=(keys := ["gvkey", "datadate"]))
        exclude_cols = cols - set([*keys, "capx", "at", NAME])
        return (
            pd.DataFrame.from_records(nparray, exclude=exclude_cols),
            VARIABLE_LABELS,
        )



"""Define the professor who's the main supervisor (process) of estimations"""

import os
from typing import List, Set, Dict
from multiprocessing import Manager, active_children
from frds.ra import ResearchAssistant
from frds.measures import Measure
from frds.data import Dataset, DataManager
from frds.typing import TableID, SharedMemoryInfo


class Professor:
    """Professor is the main supervisor (process) for the entire estimation.
    The professor dispatches jobs to `Research Assistant`s who actually perform
    the estimation.

    Parameters
    ----------
    assistants : int, optional
        The maximum concurrency, by default os.cpu_count(). Note that because
        some measures may spwan subprocesses as well. This doesn't guarantee the
        actual number of subprocesses will be no more than `assistants`.
    """

    def __init__(self, assistants=os.cpu_count(), config=None, progress=None):
        self._assistants = assistants
        self._datamanager = DataManager(config=config)
        # The professor's office is where communication happens!
        self.office = Manager().Namespace()
        # The database stores the shared memory info for each table
        self.office.database: Dict[TableID, SharedMemoryInfo] = None
        # The configuration of login info and data/result directories
        self.office.config = config
        # Function to update progress message
        self.progress = print if progress is None else progress

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._datamanager.shutdown()

    @staticmethod
    def _prepare_datasets(measures: List[Measure]) -> List[Dataset]:
        """Given a list of measures, return the consolidated datasets required

        This is necessary because some measures may require the same dataset,
        although different variables. So the professor is going to consolidate
        the required datasets and make a bulk request to the data manager.

        Parameters
        ----------
        measures : List[Measure]
            List of measures to estimate

        Returns
        -------
        List[Dataset]
            List of consolidated datasets required
        """
        # Get a se of distinct table ids
        table_ids: Set[TableID] = set(
            dta.table_id for m in measures for dta in m.datasets_required
        )
        # Init empty dicts: TableID -> Set[VarName]
        table_varnames = {table_id: set() for table_id in table_ids}
        table_datevars = {table_id: set() for table_id in table_ids}
        # Find for each table, the distinct variables required for all measures
        for dta in (dta for m in measures for dta in m.datasets_required):
            table_varnames.get(dta.table_id).update(dta.vars)
            table_datevars.get(dta.table_id).update(dta.date_vars)
        # Consolidate datasets
        datasets: List[Dataset] = []
        for table_id in table_ids:
            src, lib, table = table_id
            varnames = list(table_varnames.get(table_id))
            datevars = list(table_datevars.get(table_id))
            datasets.append(Dataset(src, lib, table, varnames, datevars))
        return datasets

    @staticmethod
    def _prioritse_measures(measures: List[Measure]) -> List[Measure]:
        """Prioritise the measures to estimate
        
        Measures using the same dataset should be clustered and scheduled to run
        together if possible. Datasets no longer needed can then be removed.

        Parameters
        ----------
        measures : List[Measure]
            List of measures to estimate

        Returns
        -------
        List[Measure]
            List of measures to estimate (potentially reordered)
        """
        # TODO: to be implemented
        return measures

    def calculate(self, measures: List[Measure]) -> None:
        """The professor is not going to caculate the measures but will allocate
        the tasks to research assistants.

        Parameters
        ----------
        measures : List[Measure]
            Measures need to estimate
        """
        self.progress("Preparing datasets...")
        datasets = self._prepare_datasets(measures)

        self.progress("Prioritse measures to estimate...")
        measures = self._prioritse_measures(measures)

        self.progress("Requesting data from the data manager...")
        with self._datamanager as data_manager:
            self.office.data = data_manager.get_datasets(datasets)
            # Professor doesn't need to to tell RAs which datasets are required
            # for each measure, because RAs can figure out themselves!
            self.progress("Starting allocating jobs to research assistant...")
            assistants = []
            while True:
                # Limit concurrency
                if len(active_children()) < self._assistants:
                    if not measures:
                        break
                    measure = measures.pop(0)
                    assistant = ResearchAssistant(measure, self.office)
                    self.progress(f"Starting RA for {measure.name}...")
                    assistant.start()
                    assistants.append(assistant)
            self.progress("Waiting for RAs to finish estimation...")
            for assistant in assistants:
                assistant.join()
        result_dir = self.office.config.get("result_dir")
        self.progress(f"Completed! Results saved in {result_dir}")



from typing import Callable, Dict
from .professor import Professor


def main(
    measures_to_estimate=[],
    gui=False,
    progress_callback: Callable = None,
    config: Dict = None,
):
    """Main entrant of frds

    Parameters
    ----------
    measures_to_estimate : list, optional
        list of measure names to estimate, by default []
    gui : bool, optional
        if in GUI mode, by default False
    progress_callback : Callable, optional
        function used to update progress message, by default None
    """
    import inspect
    import frds.measures
    from multiprocessing import (
        get_all_start_methods,
        set_start_method,
        get_start_method,
    )

    # Use 'fork' if available
    if get_start_method(allow_none=True) is None:
        if "fork" in get_all_start_methods():
            set_start_method("fork")

    # Use standard print function if not in GUI
    progress_func = print if not gui else progress_callback.emit

    # Default to estimate all measures using default parameters
    if not gui:
        measures = [
            measure()
            for _, measure in inspect.getmembers(frds.measures, inspect.isclass)
            if not inspect.isabstract(measure)
        ]
    else:
        measures = [
            measure()
            for name, measure in inspect.getmembers(
                frds.measures, inspect.isclass
            )
            if name in measures_to_estimate
        ]

    # Professor at work!
    with Professor(config=config, progress=progress_func) as prof:
        prof.calculate(measures)


if __name__ == "__main__":

    from frds import wrds_username, wrds_password, result_dir, data_dir

    config = dict(
        wrds_username=wrds_username,
        wrds_password=wrds_password,
        result_dir=str(result_dir),
        data_dir=str(data_dir),
    )

    main(config=config)



"""Some utility functions"""

import numpy as np


def filter_funda(funda: np.recarray) -> np.recarray:
    """Standard filter on the `wrds.comp.funda` dataset

    Parameters
    ----------
    funda : np.recarray
        `wrds.comp.funda` dataset

    Returns
    -------
    np.recarray
        Filtered dataset
    """
    return funda[
        np.in1d(funda.datafmt, ("STD"))
        & np.in1d(funda.indfmt, ("INDL"))
        & np.in1d(funda.popsrc, ("D"))
        & np.in1d(funda.consol, ("C"))
    ]



import sys
import os
import inspect
from typing import List
from PyQt5.QtCore import (
    Qt,
    QRunnable,
    QObject,
    QThreadPool,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QStatusBar,
    QGroupBox,
    QVBoxLayout,
    QPushButton,
    QLabel,
    QHBoxLayout,
    QGridLayout,
    QCheckBox,
    QLineEdit,
)
from PyQt5.QtGui import QIcon
from frds import wrds_username, wrds_password, data_dir, result_dir
import frds.measures
import frds.run


class Worker(QRunnable):
    """Worker thread for running background tasks."""

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()
        self.kwargs["progress_callback"] = self.signals.progress

    @pyqtSlot()
    def run(self):
        try:
            result = self.fn(*self.args, **self.kwargs)
        except Exception as exception:
            print(exception)
            self.signals.error.emit(str(exception))
        else:
            self.signals.result.emit(result)
        finally:
            self.signals.finished.emit()


class WorkerSignals(QObject):
    """
    Defines the signals available from a running worker thread.
    Supported signals are:
    finished
        No data
    error
        `Exception`
    result
        `object` data returned from processing, anything
    """

    finished = pyqtSignal()
    error = pyqtSignal(str)
    result = pyqtSignal(object)
    progress = pyqtSignal(str)


class GUI(QDialog):
    """Main GUI"""

    def __init__(self):
        super(GUI, self).__init__(parent=None)

        author = "Mingze Gao"
        my_email = "mingze.gao@sydney.edu.au"
        my_site_url = "https://mingze-gao.com/"
        github_url = "https://github.com/mgao6767/frds/"
        frds_title = "FRDS - Financial Research Data Services"
        style_name = "Fusion"
        intro_html = f"""
        <p>Estimate a collection of corporate finance metrics on one click!</p>
        <ol>
            <li>Select measures to estimate.</li>
            <li>Enter your WRDS username and password, click "Start".</li>
            <li>Output datasets will be saved in the result directory.</li>
        </ol>
        <p>Author: <a href="{my_site_url}">{author}</a> |
        Email: <a href="mailto:{my_email}">{my_email}</a> |
        Source code: <a href="{github_url}">{github_url}</a></p><hr>"""

        self.measures: List[(str, QCheckBox)] = []
        self.status_bar = QStatusBar()
        self.status_bar.showMessage("Status: Ready.")
        self.threadpool = QThreadPool()
        self.stopped = True

        # Set window title
        self.setWindowTitle(frds_title)
        script_dir = os.path.dirname(os.path.realpath(__file__))
        icon_path = os.path.join(script_dir, "favicon.ico")
        self.setWindowIcon(QIcon(icon_path))
        # Set style
        QApplication.setStyle(style_name)

        # Intro layout
        (intro_label := QLabel(intro_html)).setOpenExternalLinks(True)
        (intro_layout := QHBoxLayout()).addWidget(intro_label)

        # Control button
        self.all_measures_btn = QCheckBox("All Measures")
        self.all_measures_btn.setCheckState(Qt.Checked)
        self.start_btn = QPushButton("Start")
        (ctrl_layout := QHBoxLayout()).addWidget(self.start_btn)

        # Measure selection
        self.measure_selection = self.create_measure_selection_layout()
        # Configuration
        self.configuration = self.create_configuration_layout()

        # Main layout
        main_layout = QGridLayout()
        # - Top intro section
        main_layout.addLayout(intro_layout, 0, 0, 1, 4)
        # - Left panel for measure selection
        main_layout.addWidget(self.measure_selection, 1, 0, 2, 1)
        # - Right panel top: configuration
        main_layout.addWidget(self.configuration, 1, 1, 1, 3)
        # - Right panel bottom: controls
        main_layout.addLayout(ctrl_layout, 2, 1, 1, 3)

        (layout := QVBoxLayout()).addLayout(main_layout)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)

        # Connect
        self.start_btn.clicked.connect(self.on_start_btn_clicked)
        self.all_measures_btn.clicked.connect(self.on_all_measures_btn_clicked)

    def on_all_measures_btn_clicked(self) -> None:
        """Select and deselect all measures"""
        checked = self.all_measures_btn.isChecked()
        for _, check_box in self.measures:
            check_box.setCheckState(Qt.Checked if checked else Qt.Unchecked)

    def create_measure_selection_layout(self) -> QGroupBox:
        """Create layout for measure selection

        It also populates `self.measures`, a list of (measure_name, QCheckBox).

        Returns
        -------
        QGroupBox
            Layout for measure selection
        """
        layout = QVBoxLayout()
        layout.addWidget(self.all_measures_btn)
        for name, measure in inspect.getmembers(frds.measures, inspect.isclass):
            if not inspect.isabstract(measure):
                (check_box := QCheckBox(name)).setCheckState(Qt.Checked)
                layout.addWidget(check_box)
                self.measures.append((name, check_box))
        (measure_selection := QGroupBox("Measures")).setLayout(layout)
        return measure_selection

    def create_configuration_layout(self) -> QGroupBox:
        """Create layout for configuration

        It uses the default configuration from `config.ini`.

        Returns
        -------
        QGroupBox
            Layout for configuration
        """
        self.wrds_username_qline = QLineEdit(wrds_username)
        self.wrds_password_qline = QLineEdit(wrds_password)
        self.wrds_password_qline.setEchoMode(QLineEdit.Password)
        self.frds_data_dir = QLineEdit(str(data_dir))
        self.frds_result_dir = QLineEdit(str(result_dir))

        layout = QGridLayout()
        layout.addWidget(QLabel("WRDS Username"), 0, 0, 1, 1)
        layout.addWidget(self.wrds_username_qline, 0, 1, 1, 2)
        layout.addWidget(QLabel("WRDS Password"), 1, 0, 1, 1)
        layout.addWidget(self.wrds_password_qline, 1, 1, 1, 2)
        layout.addWidget(QLabel("Data directory"), 2, 0, 1, 1)
        layout.addWidget(self.frds_data_dir, 2, 1, 1, 2)
        layout.addWidget(QLabel("Result directory"), 3, 0, 1, 1)
        layout.addWidget(self.frds_result_dir, 3, 1, 1, 2)

        (configuration_layout := QGroupBox("Configuration")).setLayout(layout)
        return configuration_layout

    def on_start_btn_clicked(self) -> None:
        """Start running estimation"""
        self.stopped = False
        self.measure_selection.setDisabled(True)
        self.configuration.setDisabled(True)
        self.start_btn.setDisabled(True)
        self.start_btn.setText("Running")
        # TODO: modify config.ini and read config each time in frds.run.main?
        config = dict(
            wrds_username=self.wrds_username_qline.text(),
            wrds_password=self.wrds_password_qline.text(),
            data_dir=self.frds_data_dir.text(),
            result_dir=self.frds_result_dir.text(),
        )
        measures_to_estimate = [
            m for m, check_box in self.measures if check_box.isChecked()
        ]
        worker = Worker(
            frds.run.main,
            measures_to_estimate=measures_to_estimate,
            gui=True,
            config=config,
        )
        worker.signals.finished.connect(self.on_completed)
        worker.signals.progress.connect(self.update_progress)
        worker.signals.error.connect(self.update_progress)
        self.threadpool.start(worker)

    def on_completed(self) -> None:
        """Callback when estimation is completed"""
        self.measure_selection.setDisabled(False)
        self.configuration.setDisabled(False)
        self.start_btn.setDisabled(False)
        self.start_btn.setText("Start")

    def update_progress(self, msg: str) -> None:
        """Update progress"""
        self.status_bar.showMessage(msg)


if __name__ == "__main__":
    app = QApplication([])
    (gui := GUI()).show()
    sys.exit(app.exec_())



from setuptools import setup, find_namespace_packages
from setuptools.command.install import install
import pathlib
import shutil
import os

requires = ["pandas", "sqlalchemy", "psycopg2-binary", "PyQt5"]


class PostInstallCommand(install):
    def run(self):
        install.run(self)
        base_dir = str(pathlib.Path("~/frds").expanduser())
        os.makedirs(base_dir, exist_ok=True)
        os.makedirs(os.path.join(base_dir, "data"), exist_ok=True)
        os.makedirs(os.path.join(base_dir, "result"), exist_ok=True)
        if not os.path.exists(os.path.join(base_dir, "config.ini")):
            shutil.copy("frds/config.ini", base_dir)


setup(
    name="frds",
    version="0.3.1",
    description="Financial Research Data Services",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Mingze Gao",
    author_email="adrian.gao@outlook.com",
    url="https://github.com/mgao6767/frds/",
    packages=find_namespace_packages(),
    package_data={
        "": ["LICENSE", "README.md", "*.ico"],
        "frds": ["config.ini"],
    },
    install_requires=requires,
    cmdclass={"install": PostInstallCommand},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    license="MIT",
)



import os
import pandas as pd
import sqlalchemy as sa

WRDS_POSTGRES_HOST = 'wrds-pgdata.wharton.upenn.edu'
WRDS_POSTGRES_PORT = 9737
WRDS_POSTGRES_DB = 'wrds'
WRDS_CONNECT_ARGS = {'sslmode': 'require', 'application_name': 'frds'}


class NotSubscribedError(PermissionError):
    pass


class SchemaNotFoundError(FileNotFoundError):
    pass


class Connection(object):
    def __init__(self, usr, pwd, autoconnect=True):
        pguri = f'postgresql://{usr}:{pwd}@{WRDS_POSTGRES_HOST}:{WRDS_POSTGRES_PORT}/{WRDS_POSTGRES_DB}'
        self.engine = sa.create_engine(pguri, connect_args=WRDS_CONNECT_ARGS)
        if autoconnect:
            self.connect()
            self.load_library_list()

    def connect(self):
        try:
            self.connection = self.engine.connect()
        except Exception as e:
            raise e

    def close(self):
        self.connection.close()
        self.engine.dispose()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.close()

    def load_library_list(self):
        """ Load the list of Postgres schemata (c.f. SAS LIBNAMEs)
              the user has permission to access. """
        self.insp = sa.inspect(self.connection)
        print("Loading library list...")
        query = """
        WITH RECURSIVE "names"("name") AS (
            SELECT n.nspname AS "name"
                FROM pg_catalog.pg_namespace n
                WHERE n.nspname !~ '^pg_'
                    AND n.nspname <> 'information_schema')
            SELECT "name"
                FROM "names"
                WHERE pg_catalog.has_schema_privilege(
                    current_user, "name", 'USAGE') = TRUE;
        """
        cursor = self.connection.execute(query)
        self.schema_perm = [x[0] for x in cursor.fetchall()
                            if not (x[0].endswith('_old') or
                                    x[0].endswith('_all'))]
        print("Loading library list... Done")

    def __check_schema_perms(self, schema):
        """
            Check the permissions of the schema.
            Raise permissions error if user does not have access.
            Raise other error if the schema does not exist.

            Else, return True

            :param schema: Postgres schema name.
            :rtype: bool

        """
        if schema in self.schema_perm:
            return True
        else:
            if schema in self.insp.get_schema_names():
                raise NotSubscribedError(
                    "You do not have permission to access"
                    "the {} library".format(schema))
            else:
                raise SchemaNotFoundError(
                    "The {} library is not found.".format(schema))

    def list_libraries(self):
        """
            Return all the libraries (schemas) the user can access.

            :rtype: list

            Usage::
            >>> db.list_libraries()
            ['aha', 'audit', 'block', 'boardex', ...]
        """
        return self.schema_perm

    def list_tables(self, library):
        """
            Returns a list of all the views/tables/foreign tables within a schema.

            :param library: Postgres schema name.

            :rtype: list

            Usage::
            >>> db.list_tables('wrdssec')
            ['wciklink_gvkey', 'dforms', 'wciklink_cusip', 'wrds_forms', ...]
        """
        if self.__check_schema_perms(library):
            output = (self.insp.get_view_names(schema=library) +
                      self.insp.get_table_names(schema=library) +
                      self.insp.get_foreign_table_names(schema=library))
            return output

    def __get_schema_for_view(self, schema, table):
        """
        Internal function for getting the schema based on a view
        """
        sql_code = """SELECT distinct(source_ns.nspname) AS source_schema
                      FROM pg_depend
                      JOIN pg_rewrite
                        ON pg_depend.objid = pg_rewrite.oid
                      JOIN pg_class as dependent_view
                        ON pg_rewrite.ev_class = dependent_view.oid
                      JOIN pg_class as source_table
                        ON pg_depend.refobjid = source_table.oid
                      JOIN pg_attribute
                        ON pg_depend.refobjid = pg_attribute.attrelid
                          AND pg_depend.refobjsubid = pg_attribute.attnum
                      JOIN pg_namespace dependent_ns
                        ON dependent_ns.oid = dependent_view.relnamespace
                      JOIN pg_namespace source_ns
                        ON source_ns.oid = source_table.relnamespace
                      WHERE dependent_ns.nspname = '{schema}'
                        AND dependent_view.relname = '{view}';
                    """.format(schema=schema, view=table)
        if self.__check_schema_perms(schema):
            result = self.connection.execute(sql_code)
            return result.fetchone()[0]

    def describe_table(self, library, table):
        """
            Takes the library and the table and describes all the columns
              in that table.
            Includes Column Name, Column Type, Nullable?.

            :param library: Postgres schema name.
            :param table: Postgres table name.

            :rtype: pandas.DataFrame

            Usage::
            >>> db.describe_table('wrdssec_all', 'dforms')
                        name nullable     type
                  0      cik     True  VARCHAR
                  1    fdate     True     DATE
                  2  secdate     True     DATE
                  3     form     True  VARCHAR
                  4   coname     True  VARCHAR
                  5    fname     True  VARCHAR
        """
        rows = self.get_row_count(library, table)
        print("Approximately {} rows in {}.{}.".format(rows, library, table))
        table_info = pd.DataFrame.from_dict(
            self.insp.get_columns(table, schema=library))
        return table_info[['name', 'nullable', 'type']]

    def get_row_count(self, library, table):
        """
            Uses the library and table to get the approximate
              row count for the table.

            :param library: Postgres schema name.
            :param table: Postgres table name.

            :rtype: int

            Usage::
            >>> db.get_row_count('wrdssec', 'dforms')
            16378400
        """
        if 'taq' in library:
            schema = library
            print("The row count will return 0 due to the structure of TAQ")
        else:
            schema = self.__get_schema_for_view(library, table)
        if schema:
            sqlstmt = f"""
                SELECT reltuples
                  FROM pg_class r
                  JOIN pg_namespace n
                    ON (r.relnamespace = n.oid)
                  WHERE r.relkind in ('r', 'f')
                    AND n.nspname = '{schema}'
                    AND r.relname = '{table}';
                """
            try:
                result = self.connection.execute(sqlstmt)
                return int(result.fetchone()[0])
            except Exception as e:
                print(
                    "There was a problem with retrieving"
                    "the row count: {}".format(e))
                return 0
        else:
            print("There was a problem with retrieving the schema")
            return None

    def raw_sql(self, sql, coerce_float=True, date_cols=None, index_col=None, params=None):
        """
            Queries the database using a raw SQL string.

            :param sql: SQL code in string object.
            :param coerce_float: (optional) boolean, default: True
                Attempt to convert values to non-string, non-numeric objects
                to floating point. Can result in loss of precision.
            :param date_cols: (optional) list or dict, default: None
                - List of column names to parse as date
                - Dict of ``{column_name: format string}`` where
                    format string is:
                      strftime compatible in case of parsing string times or
                      is one of (D, s, ns, ms, us) in case of parsing
                        integer timestamps
                - Dict of ``{column_name: arg dict}``,
                    where the arg dict corresponds to the keyword arguments of
                      :func:`pandas.to_datetime`
            :param index_col: (optional) string or list of strings,
              default: None
                Column(s) to set as index(MultiIndex)
            :param params: parameters to SQL query, if parameterized.

            :rtype: pandas.DataFrame

            Usage ::
            # Basic Usage
            >>> data = db.raw_sql('select cik, fdate, coname from wrdssec_all.dforms;', date_cols=['fdate'], index_col='cik')
            >>> data.head()
                cik        fdate       coname
                0000000003 1995-02-15  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1996-02-14  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1997-02-19  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1998-03-02  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1998-03-10  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y..
                ...

            # Parameterized SQL query
            >>> parm = {'syms': ('A', 'AA', 'AAPL'), 'num_shares': 50000}
            >>> data = db.raw_sql('select * from taqmsec.ctm_20030910 where sym_root in %(syms)s and size > %(num_shares)s', params=parm)
            >>> data.head()
                      date           time_m ex sym_root sym_suffix tr_scond      size   price tr_stopind tr_corr     tr_seqnum tr_source tr_rf
                2003-09-10  11:02:09.485000  T        A       None     None  211400.0  25.350          N      00  1.929952e+15         C  None
                2003-09-10  11:04:29.508000  N        A       None     None   55500.0  25.180          N      00  1.929952e+15         C  None
                2003-09-10  15:08:21.155000  N        A       None     None   50500.0  24.470          N      00  1.929967e+15         C  None
                2003-09-10  16:10:35.522000  T        A       None        B   71900.0  24.918          N      00  1.929970e+15         C  None
                2003-09-10  09:35:20.709000  N       AA       None     None  108100.0  28.200          N      00  1.929947e+15         C  None
        """
        try:
            return pd.read_sql_query(
                sql,
                self.connection,
                coerce_float=coerce_float,
                parse_dates=date_cols,
                index_col=index_col,
                params=params)
        except sa.exc.ProgrammingError as e:
            raise e

    def get_table(self, library, table, obs=-1, offset=0,
                  columns=None, coerce_float=None, index_col=None,
                  date_cols=None):
        """
            Creates a data frame from an entire table in the database.

            :param sql: SQL code in string object.
            :param library: Postgres schema name.

            :param obs: (optional) int, default: -1
                Specifies the number of observations to pull from the table.
                An integer less than 0 will return the entire table.
            :param offset: (optional) int, default: 0
                Specifies the starting point for the query.
                An offset of 0 will start selecting from the beginning.
            :param columns: (optional) list or tuple, default: None
                Specifies the columns to be included in the output data frame.
            :param coerce_float: (optional) boolean, default: True
                Attempt to convert values to non-string, non-numeric objects
                to floating point. Can result in loss of precision.
            :param date_cols: (optional) list or dict, default: None
                - List of column names to parse as date
                - Dict of ``{column_name: format string}``
                    where format string is
                      strftime compatible in case of parsing string times or
                      is one of (D, s, ns, ms, us) in case of parsing
                        integer timestamps
                - Dict of ``{column_name: arg dict}``,
                    where the arg dict corresponds to the keyword arguments of
                      :func:`pandas.to_datetime`
            :param index_col: (optional) string or list of strings,
              default: None
                Column(s) to set as index(MultiIndex)

            :rtype: pandas.DataFrame

            Usage ::
            >>> data = db.get_table('wrdssec_all', 'dforms', obs=1000, columns=['cik', 'fdate', 'coname'])
            >>> data.head()
                cik        fdate       coname
                0000000003 1995-02-15  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1996-02-14  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1997-02-19  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1998-03-02  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
                0000000003 1998-03-10  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y..
                ...

        """
        if obs < 0:
            obsstmt = ''
        else:
            obsstmt = ' LIMIT {}'.format(obs)
        if columns is None:
            cols = '*'
        else:
            cols = ','.join(columns)
        if self.__check_schema_perms(library):
            sqlstmt = ('SELECT {cols} FROM {schema}.{table} {obsstmt} OFFSET {offset};'.format(
                       cols=cols,
                       schema=library,
                       table=table,
                       obsstmt=obsstmt,
                       offset=offset))
            return self.raw_sql(
                sqlstmt,
                coerce_float=coerce_float,
                index_col=index_col,
                date_cols=date_cols)



"""Define some types used in frds"""

from typing import List, Set, Tuple, NewType
from multiprocessing.shared_memory import SharedMemory
from numpy import dtype

TableID = NewType("TableID", Tuple[str, str, str])
SharedMemoryInfo = NewType(
    "SharedMemoryInfo", Tuple[SharedMemory, tuple, dtype]
)
